@@ -18,7 +18,6 @@ _socket = socket.socket
 _socketpair = socket.socketpair
 _fromfd = socket.fromfd
 
-_SOCKET_CLOSED = frozenset((errno.ECONNRESET, errno.ENOTCONN, errno.ESHUTDOWN))
 _BLOCKING_OP = frozenset((
         errno.EINPROGRESS, errno.EAGAIN, errno.EWOULDBLOCK, errno.EALREADY))
 _CANT_SEND = frozenset((errno.EWOULDBLOCK, errno.ENOTCONN))
@@ -58,10 +57,6 @@ class Socket(object):
         self._readable = util.Event()
         self._writable = util.Event()
 
-        # register the socket for this fd in the scheduler
-        scheduler._register_fd(
-                self._fileno, self._on_readable, self._on_writable)
-
     def _on_readable(self):
         self._readable.set()
         self._readable.clear()
@@ -72,29 +67,38 @@ class Socket(object):
 
     def _poller_evmask(self, poller, events):
         mask = 0
+        rd, wr = False, False
         if 'r' in events:
             mask |= poller.INMASK
+            rd = True
         if 'w' in events:
             mask |= poller.OUTMASK
+            wr = True
         if 'e' in events:
             mask |= poller.ERRMASK
-        return mask
+        return mask, rd, wr
 
     @contextlib.contextmanager
     def _registered(self, events=None):
         poller = scheduler.state.poller
         if events:
-            events = self._poller_evmask(poller, events)
+            events, rd, wr = self._poller_evmask(poller, events)
         try:
             counter = poller.register(self, events)
         except EnvironmentError, exc:
+            tb = sys.exc_info()[2]
             if exc.args and exc.args[0] in errno.errorcode:
-                raise socket.error(*exc.args)
+                raise socket.error, socket.error(*exc.args), tb
             raise
 
+        rd = self._on_readable if rd else None
+        wr = self._on_writable if wr else None
+        scheduler._register_fd(self._fileno, rd, wr)
+
         try:
             yield
         finally:
+            scheduler._unregister_fd(self._fileno, rd, wr)
             try:
                 poller.unregister(self, counter)
             except EnvironmentError, exc:
@@ -317,9 +321,6 @@ class Socket(object):
                     return self._sock.recv(bufsize, flags)
                 except socket.error, exc:
                     if not self._blocking or exc[0] not in _BLOCKING_OP:
-                        if exc[0] in _SOCKET_CLOSED:
-                            self._closed = True
-                            return ''
                         raise
                     sys.exc_clear()
                     if self._readable.wait(self.gettimeout()):
@@ -354,9 +355,6 @@ class Socket(object):
                     return self._sock.recv_into(buffer, bufsize, flags)
                 except socket.error, exc:
                     if not self._blocking or exc[0] not in _BLOCKING_OP:
-                        if exc[0] in _SOCKET_CLOSED:
-                            self._closed = True
-                            return
                         raise
                     sys.exc_clear()
                     if self._readable.wait(self.gettimeout()):
@@ -390,9 +388,6 @@ class Socket(object):
                     return self._sock.recvfrom(bufsize, flags)
                 except socket.error, exc:
                     if not self._blocking or exc[0] not in _BLOCKING_OP:
-                        if exc[0] in _SOCKET_CLOSED:
-                            self._closed = True
-                            return '', (None, 0)
                         raise
                     sys.exc_clear()
                     if self._readable.wait(self.gettimeout()):
@@ -429,9 +424,6 @@ class Socket(object):
                     return self._sock.recvfrom_into(buffer, bufsize, flags=0)
                 except socket.error, exc:
                     if not self._blocking or exc[0] not in _BLOCKING_OP:
-                        if exc[0] in _SOCKET_CLOSED:
-                            self._closed = True
-                            return '', (None, 0)
                         raise
                     sys.exc_clear()
                     if self._readable.wait(self.gettimeout()):
